/*-------------------------------------------------------------------*/
/* CBRAND.C      Mikko Malinen                                       */
/*                                                                   */
/*                                                                   */
/* Creates random codebooks by sampling from TS. Codebook size       */
/* can be given.                                                     */
/*                                                                   */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "CBRAND"
#define VersionNumber  "V.0.30"
#define LastUpdated    "21.08.12" /* MM */

/* ----------------------------------------------------------------- */
/*      Changelog:                                                   */
/*                                                                   */
/* MM     0.30  Random number generator seed setting added           */
/*                                                                   */
/* MM     0.29  Almost completely rewritten, from 21. 07. 09.        */
/*              Frequencies can be taken into account, alternative   */
/*              whole space sampling, CB Freq:s calculation,         */
/*              minor changes in output, repeats, best codebook,     */
/*              mean added. Cb routines used.                        */
/* JK     0.28  Previous program by Juha Kivijrvi                   */
/*                                                                   */
/* ----------------------------------------------------------------- */


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

#define  FACTFILE  "cbrand.fac"

#include "parametr.c"
#include "cb.h"
#include "file.h"
#include "memctrl.h"
#include "owntypes.h"
#include "interfc.h"
#include "random.h"
#include "reporting.h"
#include "rls.h"


/* ========================== PROTOTYPES ============================= */


static void PrintOperatingInfo(char  TSetName[],
                               char  CBName[]);
void PrintInfo(void);
static void ShutDown(TRAININGSET* TS, CODEBOOK* CB, CODEBOOK* CBbest, PARTITIONING* P);
char* InfoLine(void);
void PrintRepeatInfo(int quietLevel, int i, int iter, double error,
		     double time);
static void PrintFooter(int quietLevel, double error, double meanError,
                        int repeats, double totalTime);
void CreateRandomVectorWholeSpace(CODEBOOK *CB, int vectorNumber);
void CreateRandomCBWholeSpace(CODEBOOK* CB);
void UpdateFreqsToCB(CODEBOOK *CB, PARTITIONING *P);
int FreqToIndex(int freqsum, TRAININGSET* TS);
void GetRandomVector(BOOKNODE* node, TRAININGSET* TS);
void GenerateRandomCBWithFreqsIntoAccount(CODEBOOK* CB, TRAININGSET* TS);
void GenerateRandomCodebook(TRAININGSET *TS, CODEBOOK *CB, CODEBOOK *CBbest,
			    PARTITIONING *P, double *errorSumOfRepeats,
			    double *besterror, double *totalTime);


/* ========================== FUNCTIONS ============================== */


static void PrintOperatingInfo(char  TSetName[],
                               char  CBName[])
{
  if( Value(QuietLevel) >= 2 )
    {
      printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
      printf("Training Set: %s\n", TSetName);
      printf("Codebook:     %s\n", CBName);
      PrintSelectedOptions();
      printf("\n");
    }
  else if( Value(QuietLevel) >= 1 )
    {
      printf("%-12s ", TSetName);
      printf("%-12s ", CBName);
    }
}


/*-------------------------------------------------------------------*/


void PrintInfo(void)
{
  printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
  printf("This program creates a random codebook from a given training set or,\n");
  printf("alternatively, using the whole space of training set.\n");
  printf("Usage  : %s [%coption] <training-set> <codebook>\n", ProgName, OPTION_SYMBOL);
  printf("For example: %s bridge bridge-km\n", ProgName);
  printf("\n  Options:\n");
  PrintOptions();
  printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
}


/* ================================================================= */


static void ShutDown(TRAININGSET* TS, CODEBOOK* CB, CODEBOOK* CBbest, PARTITIONING* P)
{
  FreeCodebook(TS);
  FreeCodebook(CB);
  FreeCodebook(CBbest);
  FreePartitioning(P);
}


/* ================================================================= */


char* InfoLine(void)
{
  char* p;
  int len;

  len = strlen(ProgName)+strlen(VersionNumber)+strlen(LastUpdated)+4;
  p   = (char*) malloc(len*sizeof(char));

  if (!p)
    {
      ErrorMessage("ERROR: Allocating memory failed!\n");
      ExitProcessing(FATAL_ERROR);
    }

  sprintf(p, "%s\t%s\t%s", ProgName, VersionNumber, LastUpdated);

  return p;
} 


/* ================================================================= */


void PrintRepeatInfo(int quietLevel, int i, int iter, double error,
		      double time)
{
  if ((quietLevel >= 2) && (i == 0))
    {
      PrintMessage("%d\t\t%f\t%f\n", iter, error, time);
    }
} 


/* ------------------------------------------------------------------ */


static void PrintFooter(int quietLevel, double error, double meanError, int repeats, double totalTime)
{
  if (quietLevel >= 2)
    {
      PrintMessage("\naverage mse = %f  best mse = %f\ntime = %f (s)  %d repeats  \n\n", meanError, error, totalTime, repeats);
    }
  else if (quietLevel == 1)
    {
      PrintMessage("%f\t%f\n", error, totalTime);
    }
} 


/*-------------------------------------------------------------------*/



void CreateRandomVectorWholeSpace(CODEBOOK *CB, int vectorNumber)
{
  int j,k;
  double vsum;
  vsum = 0;
  for (j = 0; j<VectorSize(CB);j++) /* Through all components of vector*/
    {
      k = IRI(CB->MinValue,CB->MaxValue);
      VectorScalar(CB,vectorNumber,j) = k;
      vsum += k;
    }
  VectorMean(CB,vectorNumber) = (double) (vsum/(double)VectorSize(CB));
  VectorFreq(CB,vectorNumber) = 1;
}


/*-------------------------------------------------------------------*/


void CreateRandomCBWholeSpace(CODEBOOK* CB)
{
  int i;
  for (i=0;i<BookSize(CB);i++)    /* Through all vectors in CB */
    {
      CreateRandomVectorWholeSpace(CB, i);
    }
}


/*-------------------------------------------------------------------*/


void UpdateFreqsToCB(CODEBOOK *CB,PARTITIONING *P)
{
  /* Frequencies of vectors in CB are calculated from P */
  int vectorNumber;
  for (vectorNumber=0; vectorNumber < BookSize(CB); vectorNumber++)
    {
      VectorFreq(CB,vectorNumber) = CCFreq(P,vectorNumber);
    }
  TotalFreq(CB) = P->TSsize;
}


/*-------------------------------------------------------------------*/


int FreqToIndex(int freqsum, TRAININGSET* TS)

{
  int i=0;
  int l=0;

  while (l<freqsum)
    {
      l += VectorFreq(TS,i);
      i++;
    }
  return (i-1);
}


/*-------------------------------------------------------------------*/


void GetRandomVector(BOOKNODE* node, TRAININGSET* TS)
{
  int t;

  t=irand(1,TotalFreq(TS));
  CopyNode( &Node(TS, FreqToIndex(t,TS)), node, VectorSize(TS) );
}


/*-------------------------------------------------------------------*/


void GenerateRandomCBWithFreqsIntoAccount(CODEBOOK* CB, TRAININGSET* TS)
{
  int i;
  YESNO Ok;

  for (i=0; i<Value(Clusters); i++)
    {
    do
      {
        Ok = YES;
        GetRandomVector(&Node(CB,i), TS);

	/* Add this possibly in the future. Option AllowDuplicates is not 
           currently in use (v0.29). If this option is taken in to use, the 
           consistency to other options should be checked. */
        /*
        if (!Value(AllowDuplicates))
	  for (j=0; j<i; j++)
	    if (EqualVectors(Vector(CB,i), Vector(CB,j), VectorSize(CB)))
	      Ok=NO;
	*/
      }
    while (Ok==NO);
    }
}



void GenerateRandomCodebook(TRAININGSET *TS, CODEBOOK *CB, CODEBOOK *CBbest, 
                       PARTITIONING *P, double *errorSumOfRepeats, 
                       double *besterror, double *totalTime)
{
  double error;
  int repeats;
  *besterror = 0;  /* 0 is ok, we need this only to be initialized */
  *errorSumOfRepeats = 0;

  for (repeats = 1; repeats < Value(Repeats)+1; repeats++)
    {
      if (Value(UseWholeSpace)==0)
        {
	  if (Value(TakeFreqsIntoAccount)==0)
	    {
              SelectRandomRepresentatives(TS, CB);
	    }
	  else 
	    {
	      GenerateRandomCBWithFreqsIntoAccount(CB, TS);
	    }
        }
      else
        {
          CreateRandomCBWholeSpace(CB);
        }
      error = GenerateOptimalPartitioning(TS,CB,P);
      *errorSumOfRepeats += error;
      PrintRepeatInfo(Value(QuietLevel), 0, repeats, error,
		      GetClock(*totalTime));
      if ((repeats==1)||(error<*besterror))
        {
          CopyCodebook(CB, CBbest);
          *besterror = error;
        }
    }

}


/*===========================  M A I N  ==============================*/


int main(int argc, char* argv[])
{
  CODEBOOK      CB;
  TRAININGSET   TS;
  CODEBOOK      CBbest;
  PARTITIONING  P;
  char          TSName[80] = {'\0'};
  char          CBName[80] = {'\0'}; 
  CBFILETYPE    ft1;
  char *        genMethod;
  int           size;
  double        besterror;
  double        errorSumOfRepeats;
  double        totalTime;

  ParameterInfo paraminfo[2] =  { { TSName, "", 0, INFILE  },
                                  { CBName, "", 0, INFILE  } };

  ParseParameters( argc, argv, 2, paraminfo);

  ft1 = DetermineCBFileTypeConsideringOrder(TSName, 12);
  if ( ft1 != TSFILE || ft1 == NOTFOUND )
    {
      PrintInfo();
      exit(-1);
    }

  CheckFileName(CBName, FormatNameCB);
  ReadTrainingSet(TSName, &TS);
  genMethod = InfoLine();
  size = Value(Clusters);   
  CreateNewCodebook(&CB, size, &TS);  
  AddGenerationMethod(&CB, genMethod);
  CreateNewCodebook(&CBbest, size, &TS);
  CreateNewPartitioning(&P, &TS, size);
  PrintOperatingInfo(TSName, CBName);
  SetClock(&totalTime);

  initrandom(0);

  GenerateRandomCodebook(&TS, &CB, &CBbest, &P, &errorSumOfRepeats, 
                         &besterror, &totalTime);

  PrintFooter(Value(QuietLevel), besterror, errorSumOfRepeats/(Value(Repeats)), Value(Repeats), GetClock(totalTime));

  UpdateFreqsToCB(&CBbest, &P);
  WriteCodebook(CBName, &CBbest, Value(OverWrite));

  free(genMethod);
  ShutDown(&TS, &CB, &CBbest, &P);
  checkmemory();
  return( 0 );
}
